#include <stdio.h>      
#include <stdlib.h>     
#include <unistd.h>
#include <getopt.h>  
#include <QtWidgets/QApplication>
#include <QtWidgets/QMainWindow>
#include <QtWidgets/QMenuBar>   
#include <QtWidgets/QMenu>   
#include <QtWidgets/QInputDialog>   
#include <QtCore/QString>   
#include "model.h"
#include "gui.h"
#include "jackprocessor.h"

static struct option options[] =
        {{"poly", 4, 0, 'p'}, 
         {"verbose", 0, 0, 'v'},  
         {"help", 0, 0, 'h'},  
         {0, 0, 0, 0}};

int main(int argc, char *argv[])  
{
  int getopt_return;
  int option_index;
  bool verbose, ok, set_poly;
  int poly, rate, bufsize, i;    
  QFile f;
  QString qs;
  char buf[256];

  poly = 4;
  rate = 44100;
  bufsize = 1024;
  verbose = false;
  set_poly = true;
  QApplication *myApp = new QApplication(argc, argv);
  while((getopt_return = getopt_long(argc, argv, "vhp:", options, &option_index)) >= 0) {
    switch(getopt_return) {
      case 'p':
        poly = atoi(optarg);
        if (poly < 1) poly = 1; 
        if (poly > MAX_POLY) poly = MAX_POLY;
        set_poly = false;
        break;
      case 'v':
        verbose = true;
        break;
      case 'h': 
        printf("\n\nAdd64  v.3.7.4-1   (c)2011-2019 by Matthias Nagorni\n\n");
        printf("--poly            number of voices [1]\n");
        printf("--help            print this help page\n\n");
        exit(EXIT_SUCCESS);
    }
  }                          
  QMainWindow *top = new QMainWindow();
  top->setWindowTitle("Add64  v.3.7.4-1   (c)2011-2019 by Matthias Nagorni");
  f.setFileName("Add64-MIDIconfig");
  if (f.exists()) { 
    f.open(QIODevice::ReadOnly);
    f.readLine(buf, sizeof(buf));
    f.close();
    qs = QString(buf);
    poly = qs.toInt();
    set_poly = (poly == 0);
    if (poly < 1) poly = 1;
    if (poly > MAX_POLY) poly = MAX_POLY;
  }                                                                    
  if (set_poly) {
    i = QInputDialog::getInt(top, "Add64 Polyphony", "Polyphony: ", 4, 1, MAX_POLY, 1, &ok);
    if (ok && (i > 0) && (i <= MAX_POLY)) {
      poly = i;
    } 
  }   
  Model model(poly);
  model.setVerbose(verbose);
  model.setRate(rate);
  model.setBufSize(bufsize);
  model.setEditRange(0, poly - 1);
  Gui gui(&model, top);
  QMenu *fileMenu = top->menuBar()->addMenu("&File");
  fileMenu->addAction("&Load Sound (XML Format)", &gui, SLOT(loadXML()));
  fileMenu->addAction("&Save Sound (XML Format)", &gui, SLOT(saveXML()));
  fileMenu->addSeparator();
  fileMenu->addAction("Loa&d MIDI Map", &gui, SLOT(loadMidi()));
  fileMenu->addAction("Sav&e MIDI Map", &gui, SLOT(saveMidi()));
  fileMenu->addSeparator();
  fileMenu->addAction("&Quit", myApp, SLOT(quit()));
  QMenu *commentMenu = top->menuBar()->addMenu("&Comment");
  commentMenu->addAction("&New Comment", &gui, SLOT(addComment()));
  commentMenu->addAction("&Edit Comment", &gui, SLOT(editComment()));
  commentMenu->addAction("&Show Comment", &gui, SLOT(showComment()));
  QMenu *scalaMenu = top->menuBar()->addMenu("&Scala");
  scalaMenu->addAction("&Edit Scala Settings", &gui, SLOT(showScalaDialog()));
  QMenu *helpMenu = top->menuBar()->addMenu("&Help");
  helpMenu->addAction("&About Add64", &gui, SLOT(showAbout()));
  helpMenu->addAction("&About Qt", myApp, SLOT(aboutQt()));
  QObject::connect(myApp, SIGNAL(aboutToQuit()), &gui, SLOT(closeGui()));
  top->setCentralWidget(&gui);
  top->show();
  return myApp->exec(); 
}
